@@ -2,13 +2,13 @@
 
 import abc
 import ast
+import enum
 import glob
 import json
 import os
 import os.path
 import re
 import shutil
-import struct
 import sys
 
 import charliecloud as ch
@@ -166,6 +166,14 @@ def main(cli_):
    global image_ct
    image_ct = sum(1 for i in tree.children_("from_"))
 
+   # If we use RSYNC, error out quickly if appropriate rsync(1) not present.
+   if (tree.child("rsync") is not None):
+      try:
+         ch.version_check(["rsync", "--version"], ch.RSYNC_MIN)
+      except ch.Fatal_Error:
+         ch.ERROR("Dockerfile uses RSYNC, so rsync(1) is required")
+         raise
+
    # Traverse the tree and do what it says.
    #
    # We don’t actually care whether the tree is traversed breadth-first or
@@ -206,7 +214,7 @@ def main(cli_):
            % (ml.instruction_total_ct, ml.inst_prev.image))
    # FIXME: remove when we’re done encouraging people to use the build cache.
    if (isinstance(bu.cache, bu.Disabled_Cache)):
-      ch.INFO("build slow? consider enabling the new build cache",
+      ch.INFO("build slow? consider enabling the build cache",
               "https://hpc.github.io/charliecloud/command-usage.html#build-cache")
 
 
@@ -242,6 +250,7 @@ class Main_Loop(lark.Visitor):
             return
          except ch.Fatal_Error:
             inst.announce_maybe()
+            inst.prepare_rollback()
             raise
          if (inst.miss):
             if (self.miss_ct == 1):
@@ -391,7 +400,10 @@ class Instruction(abc.ABC):
    def announce_maybe(self):
       "Announce myself if I haven’t already been announced."
       if (not self.announced_p):
-         ch.INFO("%3s%s %s" % (self.lineno, self.status_char, self))
+         self_ = str(self)
+         if (ch.user() == "qwofford" and sys.stderr.isatty()):
+            self_ = re.sub(r"^RSYNC", "NSYNC", self_)
+         ch.INFO("%3s%s %s" % (self.lineno, self.status_char, self_))
          self.announced_p = True
 
    def chdir(self, path):
@@ -413,9 +425,6 @@ class Instruction(abc.ABC):
       self.git_hash = bu.cache.commit(path, self.sid, str(self),
                                       self.commit_files)
 
-   def ready(self):
-      bu.cache.ready(self.image)
-
    def execute(self):
       """Do what the instruction says. At this point, the unpack directory is
          all ready to go. Thus, the method is cache-ignorant."""
@@ -469,10 +478,10 @@ class Instruction(abc.ABC):
               time in prepare() should call announce_maybe() as soon as they
               know hit/miss status.
 
-           2. Errors: Calling ch.FATAL() normally exits immediately, but here
-              this often happens before the instruction has been announced
-              (see issue #1486). Therefore, the caller catches Fatal_Error,
-              announces, and then re-raises.
+           2. Errors: The caller catches Fatal_Error, announces, calls
+              prepare_rollback(), and then re-raises. This to ensure the
+              instruction is announced (see #1486) and any
+              possibly-inconsistent state is fixed before existing.
 
            3. Modifying image metadata: Instructions like ARG, ENV, FROM,
               LABEL, SHELL, and WORKDIR must modify metadata here, not in
@@ -482,6 +491,12 @@ class Instruction(abc.ABC):
       self.git_hash = bu.cache.find_sid(self.sid, self.image.ref.for_path)
       return miss_ct + int(self.miss)
 
+   def prepare_rollback(self):
+      pass  # typically a no-op
+
+   def ready(self):
+      bu.cache.ready(self.image)
+
    def rollback(self):
       """Discard everything done by execute(), which may have completed
          partially, fully, or not at all."""
@@ -490,14 +505,14 @@ class Instruction(abc.ABC):
    def unsupported_forever_warn(self, msg):
       ch.WARNING("not supported, ignored: %s %s" % (self.str_name, msg))
 
-   def unsupported_yet_warn(self, msg, issue_no):
-      ch.WARNING("not yet supported, ignored: issue #%d: %s %s"
-                 % (issue_no, self.str_name, msg))
-
    def unsupported_yet_fatal(self, msg, issue_no):
       ch.FATAL("not yet supported: issue #%d: %s %s"
                % (issue_no, self.str_name, msg))
 
+   def unsupported_yet_warn(self, msg, issue_no):
+      ch.WARNING("not yet supported, ignored: issue #%d: %s %s"
+                 % (issue_no, self.str_name, msg))
+
 
 class Instruction_Unsupported(Instruction):
 
@@ -549,6 +564,84 @@ class Instruction_No_Image(Instruction):
       return miss_ct + int(self.miss)
 
 
+class Copy(Instruction):
+
+   # Superclass for instructions that do some flavor of file copying (ADD,
+   # COPY, RSYNC).
+
+   __slots__ = ("dst",          # string b/c trailing slash is significant
+                "dst_raw",
+                "from_",
+                "src_metadata",
+                "srcs",         # strings b/c trailing slashes are significant
+                "srcs_base",
+                "srcs_raw")
+
+   @property
+   def sid_input(self):
+      return super().sid_input + self.src_metadata
+
+   def expand_dest(self):
+      """Set self.dst from self.dst_raw with environment variables expanded
+         and image root prepended."""
+      dst_raw = ch.variables_sub(self.dst_raw, self.env_build)
+      if (len(dst_raw) < 1):
+         ch.FATAL("destination is empty after expansion: %s" % self.dst_raw)
+      base = self.image.unpack_path
+      if (dst_raw[0] != "/"):
+         base //= self.workdir
+      self.dst = base // ch.variables_sub(self.dst_raw, self.env_build)
+
+   def expand_sources(self):
+      """Set self.srcs from self.srcs_raw with environment variables and globs
+         expanded, absolute paths with appropriate base, and validate that
+         they are within the sources base."""
+      if (cli.context == "-" and self.from_ is None):
+         ch.FATAL("no context because “-” given")
+      if (len(self.srcs_raw) < 1):
+         ch.FATAL("source or destination missing")
+      self.srcs_base_set()
+      self.srcs = list()
+      for src in (ch.variables_sub(i, self.env_build) for i in self.srcs_raw):
+         # glob can’t take Path
+         matches = sorted(fs.Path(i)
+                          for i in glob.glob("%s/%s" % (self.srcs_base, src)))
+         if (len(matches) == 0):
+            ch.FATAL("source not found: %s" % src)
+         for m in matches:
+            self.srcs.append(m)
+            ch.VERBOSE("source: %s" % m)
+            # Validate source is within context directory. (We need the source
+            # as given later, so don’t canonicalize persistently.) There is no
+            # clear subsitute for commonpath() in pathlib.
+            mc = m.resolve()
+            if (not os.path.commonpath([mc, self.srcs_base])
+                           .startswith(self.srcs_base)):
+               ch.FATAL("can’t copy from outside context: %s" % src)
+
+   def srcs_base_set(self):
+      "Set self.srcs_base according to context and --from."
+      if (self.from_ is None):
+         self.srcs_base = cli.context
+      else:
+         if (self.from_ == self.image_i or self.from_ == self.image_alias):
+            ch.FATAL("--from: stage %s is the current stage" % self.from_)
+         if (not self.from_ in images):
+            # FIXME: Would be nice to also report if a named stage is below.
+            if (isinstance(self.from_, int) and self.from_ < image_ct):
+               if (self.from_ < 0):
+                  ch.FATAL("--from: invalid negative stage index %d"
+                           % self.from_)
+               else:
+                  ch.FATAL("--from: stage %d does not exist yet"
+                           % self.from_)
+            else:
+               ch.FATAL("--from: stage %s does not exist" % self.from_)
+         self.srcs_base = images[self.from_].unpack_path
+      self.srcs_base = os.path.realpath(self.srcs_base)
+      ch.VERBOSE("context: %s" % self.srcs_base)
+
+
 class Arg(Instruction):
 
    __slots__ = ("key",
@@ -654,23 +747,19 @@ class I_arg_first_equals(Arg_First):
       return v
 
 
-class I_copy(Instruction):
+class I_copy(Copy):
 
    # ABANDON ALL HOPE YE WHO ENTER HERE
    #
    # Note: The Dockerfile specification for COPY is complex, messy,
    # inexplicably different from cp(1), and incomplete. We try to be
-   # bug-compatible with Docker but probably are not 100%. See the FAQ.
+   # bug-compatible with Docker (legacy builder, not BuildKit -- yes, they are
+   # different) but probably are not 100%. See the FAQ.
    #
    # Because of these weird semantics, none of this abstracted into a general
    # copy function. I don’t want people calling it except from here.
 
-   __slots__ = ("dst",
-                "dst_raw",
-                "from_",
-                "src_metadata",
-                "srcs",
-                "srcs_raw")
+   __slots__ = ()
 
    def __init__(self, *args):
       super().__init__(*args)
@@ -692,10 +781,6 @@ class I_copy(Instruction):
       self.srcs_raw = args[:-1]
       self.dst_raw = args[-1]
 
-   @property
-   def sid_input(self):
-      return super().sid_input + self.src_metadata
-
    @property
    def str_(self):
       dst = repr(self.dst) if hasattr(self, "dst") else self.dst_raw
@@ -763,7 +848,7 @@ class I_copy(Instruction):
                   dst_path.rmtree()
                else:
                   dst_path.unlink_()
-            ch.copy2(src_path, dst_path, follow_symlinks=False)
+            src_path.copy(dst_path)
 
    def copy_src_file(self, src, dst):
       """Copy file src to dst. src might be a symlink, but dst is a canonical
@@ -785,8 +870,11 @@ class I_copy(Instruction):
       assert (not dst.is_symlink())
       assert (   (dst.exists() and (dst.is_dir() or dst.is_file()))
               or (not dst.exists() and dst.parent.is_dir()))
+      if (dst.is_dir()):
+         dst //= src.name
+      src = src.resolve()
       ch.DEBUG("copying named file: %s -> %s" % (src, dst))
-      ch.copy2(src, dst, follow_symlinks=True)
+      src.copy(dst)
 
    def dest_realpath(self, unpack_path, dst):
       """Return the canonicalized version of path dst within (canonical) image
@@ -859,71 +947,16 @@ class I_copy(Instruction):
             ch.FATAL("can’t COPY: unknown file type: %s" % src)
 
    def prepare(self, miss_ct):
-      def stat_bytes(path, links=False):
-         st = path.stat_(links)
-         return (  str(path).encode("UTF-8")
-                 + struct.pack("=HQQ", st.st_mode, st.st_size, st.st_mtime_ns))
-      # Error checking.
-      if (cli.context == "-" and self.from_ is None):
-         ch.FATAL("no context because “-” given")
-      if (len(self.srcs_raw) < 1):
-         ch.FATAL("must specify at least one source")
       # Complain about unsupported stuff.
       if (self.options.pop("chown", False)):
          self.unsupported_forever_warn("--chown")
       # Any remaining options are invalid.
       self.options_assert_empty()
-      # Find the context directory.
-      if (self.from_ is None):
-         context = cli.context
-      else:
-         if (self.from_ == self.image_i or self.from_ == self.image_alias):
-            ch.FATAL("--from: stage %s is the current stage" % self.from_)
-         if (not self.from_ in images):
-            # FIXME: Would be nice to also report if a named stage is below.
-            if (isinstance(self.from_, int) and self.from_ < image_ct):
-               if (self.from_ < 0):
-                  ch.FATAL("--from: invalid negative stage index %d"
-                           % self.from_)
-               else:
-                  ch.FATAL("--from: stage %d does not exist yet"
-                           % self.from_)
-            else:
-               ch.FATAL("--from: stage %s does not exist" % self.from_)
-         context = images[self.from_].unpack_path
-      context_canon = os.path.realpath(context)
-      ch.VERBOSE("context: %s" % context)
-      # Expand sources.
-      self.srcs = list()
-      for src in (ch.variables_sub(i, self.env_build) for i in self.srcs_raw):
-         # glob can’t take Path
-         matches = [fs.Path(i) for i in glob.glob("%s/%s" % (context, src))]
-         if (len(matches) == 0):
-            ch.FATAL("source file not found: %s" % src)
-         for i in matches:
-            self.srcs.append(i)
-            ch.VERBOSE("source: %s" % i)
-      # Expand destination.
+      # Expand operands.
+      self.expand_sources()
       self.dst = ch.variables_sub(self.dst_raw, self.env_build)
-      # Validate sources are within context directory. (Can’t convert to
-      # canonical paths yet because we need the source path as given.)
-      for src in self.srcs:
-         src_canon = src.resolve()
-         if (not os.path.commonpath([src_canon, context_canon])
-                 .startswith(context_canon)): # no clear substitute for
-                                              # commonpath in pathlib
-            ch.FATAL("can’t copy from outside context: %s" % src)
       # Gather metadata for hashing.
-      # FIXME: Locale issues related to sorting?
-      self.src_metadata = bytearray()
-      for src in self.srcs:
-         self.src_metadata += stat_bytes(src, links=True)
-         if (src.is_dir()):
-            for (dir_, dirs, files) in ch.walk(src):
-               self.src_metadata += stat_bytes(dir_)
-               for f in sorted(files):
-                  self.src_metadata += stat_bytes(dir_ // f)
-               dirs.sort()
+      self.src_metadata = fs.Path.stat_bytes_all(self.srcs)
       # Pass on to superclass.
       return super().prepare(miss_ct)
 
@@ -1059,6 +1092,10 @@ class I_from_(Instruction):
       assert (isinstance(bu.cache, bu.Disabled_Cache))
       super().checkout_for_build(self.base_image)
 
+   def execute(self):
+      # Everything happens in prepare().
+      pass
+
    def metadata_update(self, *args):
       # FROM doesn’t update metadata because it never misses when the cache is
       # enabled, so this would never be called, and we want disabled results
@@ -1142,9 +1179,155 @@ class I_from_(Instruction):
       # Done.
       return int(self.miss)  # will still miss in disabled mode
 
+   def prepare_rollback(self):
+      # AFAICT the only thing that might be busted is the unpack directories
+      # for either the base image or the image. We could probably be smarter
+      # about this, but for now just delete them.
+      try:
+         base_image = self.base_image
+      except AttributeError:
+         base_image = None
+      try:
+         image = self.image
+      except AttributeError:
+         image = None
+      if (base_image is not None or image is not None):
+         ch.INFO("something went wrong, rolling back ...")
+         if (base_image is not None):
+            bu.cache.unpack_delete(base_image, missing_ok=True)
+         if (image is not None):
+            bu.cache.unpack_delete(image, missing_ok=True)
+
+
+class I_rsync(Copy):
+
+   __slots__ = ("plus_option",
+                "rsync_options")
+
+   def __init__(self, *args):
+      super().__init__(*args)
+      self.from_ = None  # not supported yet
+      line_no = self.tree.line
+      st = self.tree.child("option_plus")
+      self.plus_option = "l" if st is None else st.terminal("OPTION_LETTER")
+      options_done = False
+      self.rsync_options = list()
+      self.srcs_raw = list()
+      for word in self.tree.terminals("WORDE"):
+         if (not options_done and word.startswith("-")):
+            # Option. See assumption in docs that makes parsing a lot easier.
+            if (word == "--"):             # end of options
+               options_done = True
+            elif (word.startswith("--")):  # long option
+               self.rsync_options.append(word)
+            else:                          # short option(s)
+               if (len(word) == 1):
+                  ch.FATAL("RSYNC: %d: invalid argument: %s" % (line_no, word))
+               # Append options individually so we can process them more later.
+               for m in re.finditer(r"[^=]=.*$|[^=]", word[1:]):
+                  self.rsync_options.append("-" + m[0])
+            continue
+         # Not an option, so it must be a source or destination path.
+         self.srcs_raw.append(word)
+      if (len(self.srcs_raw) == 0):
+         ch.FATAL("RSYNC: %d: source and destination missing" % line_no)
+      self.dst_raw = self.srcs_raw.pop()
+
+   @property
+   def rsync_options_concise(self):
+      "Return self.rsync_options with short options coalesced."
+      # We don’t group short options with an argument even though we could
+      # because it seems confusing, e.g. “-ab=c” vs. “-a -b=c”.
+      def ship_out():
+         nonlocal group
+         if (group != ""):
+            ret.append(group)
+            group = ""
+      ret = list()
+      group = ""
+      for o in self.rsync_options:
+         if (o.startswith("--")):  # long option, not grouped
+            ship_out()
+            ret.append(o)
+         elif (len(o) > 2):        # short option with argument, not grouped
+            ship_out()
+            ret.append(o)
+         else:                     # short option without argument, grouped
+            if (group == ""):
+               group = "-"
+            group += o[1:]         # add to group
+      ship_out()
+      return ret
+
+   @property
+   def str_(self):
+      ret = list()
+      if (self.plus_option is not None):
+         ret.append("+" + self.plus_option)
+      if (len(self.rsync_options_concise) > 0):
+         ret += self.rsync_options_concise
+      ret += self.srcs_raw
+      ret.append(self.dst_raw)
+      return " ".join(ret)
+
    def execute(self):
-      # Everything happens in prepare().
-      pass
+      plus_options = list()
+      if (self.plus_option in "lmu"):  # no action needed for +z
+         # see man page for explanations
+         plus_options = ["-@=-1", "-AHSXpr"]
+         if (sys.stderr.isatty()):
+            plus_options += ["--info=progress2"]
+         if (self.plus_option == "l"):
+            plus_options += ["-l", "--safe-links"]
+         elif (self.plus_option == "u"):
+            plus_options += ["-l", "--copy-unsafe-links"]
+      ch.cmd(["rsync"] + plus_options + self.rsync_options_concise
+                       + self.srcs + [self.dst])
+
+   def expand_rsync_froms(self):
+      for i in range(len(self.rsync_options)):
+         o = self.rsync_options[i]
+         m = re.search("^--([a-z]+)-from=(.+)$", o)
+         if (m is not None):
+            key = m[1]
+            if (m[2] == "-"):
+               ch.FATAL("--*-from: can’t use standard input")
+            elif (":" in m[2]):
+               ch.FATAL("--*-from: can’t use remote hosts (colon in path)")
+            path = ch.Path(m[2])
+            if (path.is_absolute()):
+               path = self.image.unpack_path // path
+            else:
+               path = self.srcs_base // path
+            self.rsync_options[i] = "--%s-from=%s" % (key, path)
+
+   def prepare(self, miss_ct):
+      self.rsync_validate()
+      # Expand operands.
+      self.expand_sources()
+      self.expand_dest()
+      self.expand_rsync_froms()
+      # Gather metadata for hashing.
+      self.src_metadata = fs.Path.stat_bytes_all(self.srcs)
+      # Pass on to superclass.
+      return super().prepare(miss_ct)
+
+   def rsync_validate(self):
+      # Reject bad + options.
+      if (self.plus_option not in ("mluz")):
+         ch.FATAL("invalid plus option: %s" % self.plus_option)
+      # Reject SSH and rsync transports. I *believe* simply the presence of
+      # “:” (colon) in the filename triggers this behavior.
+      for src in self.srcs_raw:
+         if (":" in src):
+            ch.FATAL("SSH and rsync transports not supported: %s" % src)
+      # Reject bad flags.
+      bad = { "--daemon",
+              "-n", "--dry-run",
+              "--remove-source-files" }
+      for o in self.rsync_options:
+         if (o in bad):
+            ch.FATAL("disallowed option: %s" % o)
 
 
 class Run(Instruction):
@@ -1168,7 +1351,7 @@ class Run(Instruction):
       elif (cli.force == ch.Force_Mode.SECCOMP):
          tag = ".S"
       else:
-         assert False, "unreachable code reached"
+         assert False, "unreachable code reached (force mode = %s)" % cli.force
       return super().str_name + tag
 
    def execute(self):
@@ -1288,25 +1471,8 @@ class Environment:
    """The state we are in: environment variables, working directory, etc. Most
       of this is just passed through from the image metadata."""
 
-   # FIXME:
-   # - problem:
-   #   1. COPY (at least) needs a valid build environment to figure out if it’s
-   #      a hit or miss, which happens in prepare()
-   #   2. no files from the image are available in prepare(), so we can’t read
-   #      image metadata then
-   #      - could get it from Git if needed, but that seems complicated
-   # - valid during prepare() and execute() but not __init__()
-   #   - in particular, don’t ch.variables_sub() in __init__()
-   # - instructions that update it need to change the env object in prepare()
-   #   - WORKDIR SHELL ARG ENV
-   #   - FROM
-   #     - global images and image_i makes this harder because we need to read
-   #       the metadata of image_i - 1
-   #       - solution: remove those two globals? instructions grow image and
-   #         image_i attributes?
 
-
-## Supporting functions ###
+## Supporting functions ##
 
 def unescape(sl):
    # FIXME: This is also ugly and should go in the grammar.
@@ -1320,6 +1486,3 @@ def unescape(sl):
       sl = '"%s"' % sl
    assert (len(sl) >= 2 and sl[0] == '"' and sl[-1] == '"' and sl[-2:] != '\\"')
    return ast.literal_eval(sl)
-
-
-#  LocalWords:  earley topdown iter lineno sid keypair dst srcs pathlib
