@@ -276,8 +276,13 @@ class File_Metadata:
       self.large_name = None
       self.xattrs = dict()
       if ch.save_xattrs:
-         for xattr in os.listxattr(self.path_abs, follow_symlinks=False):
-            self.xattrs[xattr] = os.getxattr(self.path_abs, xattr, follow_symlinks=False)
+         for xattr in ch.ossafe(os.listxattr,
+                                "can’t list xattrs: %s" % self.path_abs,
+                                self.path_abs, follow_symlinks=False):
+            self.xattrs[xattr] = \
+               ch.ossafe(os.getxattr, ("can’t get xattr: %s: %s"
+                                       % (self.path_abs, xattr)),
+                         self.path_abs, xattr, follow_symlinks=False)
 
    def __getstate__(self):
       return { a:v for (a,v) in self.__dict__.items()
@@ -481,10 +486,8 @@ class File_Metadata:
          ch.ossafe(os.mkfifo, "can’t make FIFO: %s" % self.path, self.path_abs)
       elif (self.path.git_incompatible_p):
          self.path_abs.git_escaped.rename_(self.path_abs)
-      if ch.save_xattrs:
-         for (xattr, val) in self.xattrs.items():
-            ch.ossafe(os.setxattr, "unable to restore xattr: %s" % xattr,
-                      self.path_abs, xattr, val, follow_symlinks=False)
+      for (xattr, val) in self.xattrs.items():
+         self.path_abs.setxattr(xattr, val, follow_symlinks=False)
       # Recurse children.
       if (len(self.children) > 0):
          for child in self.children.values():
@@ -535,10 +538,10 @@ class File_Metadata:
       return large_name
 
    def large_restore(self):
-      "Hard link my file to the copy in large file storage."
+      "Restore large file from OOB storage."
       target = ch.storage.build_large_path(self.large_name)
-      ch.DEBUG("large file: %s: linking: %s" % (self.path_abs, self.large_name))
-      self.path_abs.hardlink(target)
+      ch.DEBUG("large file: %s: copying: %s" % (self.path_abs, self.large_name))
+      fs.copy(target, self.path_abs)
 
    def pickle(self):
       (self.image_root // PICKLE_PATH) \
@@ -1066,10 +1069,10 @@ class Enabled_Cache:
          # a young hen, especially one less than one year old
          pullet = pull.Image_Puller(img, src_ref)
          pullet.download()
-      self.worktree_add(img, "root")
       pullet.unpack(last_layer)
       sid = self.sid_from_parent(self.root_id, pullet.sid_input)
       pullet.done()
+      self.worktree_adopt(img, "root")
       commit = self.commit(img.unpack_path, sid, "PULL %s" % src_ref, [])
       self.ready(img)
       if (img.ref != src_ref):
@@ -1169,7 +1172,7 @@ class Enabled_Cache:
       print("internal files: %5d %s" % (file_ct, file_suffix))
       print("disk used:      %5d %s" % (byte_ct, byte_suffix))
       # some information directly from Git
-      if (ch.verbose >= 1):
+      if (ch.log_level >= ch.Log_Level.VERBOSE):
          out = self.git(["count-objects", "-vH"]).stdout
          print("Git statistics:")
          print(textwrap.indent(out, "  "), end="")
@@ -1181,22 +1184,6 @@ class Enabled_Cache:
       """Delete specified git tag. Used for recovering deleted branches."""
       return self.git(["tag", "-d", "&%s" % tag], *args, **kwargs)
 
-   def tree_print(self):
-      # Note the percent codes are interpreted by Git.
-      # See: https://git-scm.com/docs/git-log#_pretty_formats
-      args = ["log", "--graph", "--all", "--reflog", "--topo-order"]
-      if (ch.verbose == 0):
-         # ref names, subject (instruction), branch heads.
-         fmt = "%C(auto)%d %Creset%<|(77,trunc)%s"
-         args.append("--decorate-refs=refs/heads")
-      else:
-         # ref names, short commit hash, subject (instruction), body (state ID)
-         # FIXME: The body contains a trailing newline I can’t figure out how
-         # to remove.
-         fmt = "%C(auto)%d%C(yellow) %h %Creset%s %b"
-      self.git(args + ["--format=%s" % fmt], quiet=False)
-      print()  # blank line to separate from summary
-
    def tree_dot(self):
       have_dot()
       path_gv = fs.Path(dot_base + ".gv")
@@ -1220,12 +1207,30 @@ class Enabled_Cache:
       ch.VERBOSE("writing %s" % path_pdf)
       ch.cmd_quiet(["dot", "-Tpdf", "-o%s" % path_pdf, str(path_gv)])
 
-   def unpack_delete(self, image):
+   def tree_print(self):
+      # Note the percent codes are interpreted by Git.
+      # See: https://git-scm.com/docs/git-log#_pretty_formats
+      args = ["log", "--graph", "--all", "--reflog", "--topo-order"]
+      if (ch.log_level == ch.Log_Level.INFO):
+         # ref names, subject (instruction), branch heads.
+         fmt = "%C(auto)%d %Creset%<|(77,trunc)%s"
+         args.append("--decorate-refs=refs/heads")
+      else:
+         # ref names, short commit hash, subject (instruction), body (state ID)
+         # FIXME: The body contains a trailing newline I can’t figure out how
+         # to remove.
+         fmt = "%C(auto)%d%C(yellow) %h %Creset%s %b"
+      self.git(args + ["--format=%s" % fmt], quiet=False)
+      print()  # blank line to separate from summary
+
+   def unpack_delete(self, image, missing_ok=False):
       """Wrapper for Image.unpack_delete() that first detaches the work tree's
          head. If we delete an image's unpack path without first detaching HEAD,
          the corresponding work tree must also be deleted before the bucache
          branch. This involves multiple calls to worktrees_fix(), which is
          clunky, so we use this method instead."""
+      if (not image.unpack_exist_p and missing_ok):
+         return
       (_, commit) = self.find_commit(image.ref.for_path)
       if (commit is not None):
          # Off with her head!
@@ -1280,6 +1285,14 @@ class Enabled_Cache:
       image.unpack_path.rmtree()
       ch.storage.image_tmp.rename_(image.unpack_path)
 
+   def worktree_head(self, image):
+      cp = self.git(["rev-parse", "--short", "HEAD"],
+                    fail_ok=True, cwd=image.unpack_path)
+      if (cp.returncode != 0):
+         return None
+      else:
+         return cp.stdout.strip()
+
    def worktrees_fix(self):
       """Git stores pointers (paths) both from the main repository to each
          worktree, and in the other direction from each worktree back to the
@@ -1328,14 +1341,6 @@ class Enabled_Cache:
             ch.VERBOSE("fixed %d worktrees" % len(wt_actuals))
       t.log("re-linked worktrees")
 
-   def worktree_head(self, image):
-      cp = self.git(["rev-parse", "--short", "HEAD"],
-                    fail_ok=True, cwd=image.unpack_path)
-      if (cp.returncode != 0):
-         return None
-      else:
-         return cp.stdout.strip()
-
 
 class Rebuild_Cache(Enabled_Cache):
 
@@ -1381,7 +1386,6 @@ class Disabled_Cache(Rebuild_Cache):
          for i in itertools.chain(subdirs, files):
             (dir_ // i).chmod_min()
 
-
    def pull_lazy(self, img, src_ref, last_layer=None, pullet=None):
       if (pullet is None and os.path.exists(img.unpack_path)):
          ch.VERBOSE("base image already exists, skipping pull")
