@@ -5,6 +5,7 @@ import collections.abc
 import cProfile
 import datetime
 import enum
+import functools
 import hashlib
 import io
 import locale
@@ -46,9 +47,26 @@ class Download_Mode(enum.Enum):
 
 # Root emulation mode
 class Force_Mode(enum.Enum):
-   FAKEROOT="fakeroot"
-   SECCOMP="seccomp"
-   NONE="none"
+   FAKEROOT = "fakeroot"
+   SECCOMP = "seccomp"
+   NONE = "none"
+
+# Log level
+@functools.total_ordering
+class Log_Level(enum.Enum):
+   TRACE = 3
+   DEBUG = 2
+   VERBOSE = 1
+   INFO = 0
+   WARNING = -1
+   STDERR = -2
+   QUIET_STDERR = -3
+   # To support comparisons, we need to define at least one “ordering”
+   # operator. See: https://stackoverflow.com/a/39269589
+   def __lt__(self, other):
+      if self.__class__ is other.__class__:
+         return self.value < other.value
+      return NotImplemented
 
 
 ## Constants ##
@@ -63,8 +81,7 @@ class Force_Mode(enum.Enum):
 #
 # [1]: https://stackoverflow.com/a/45125525
 # [2]: https://github.com/docker-library/bashbrew/blob/v0.1.0/vendor/github.com/docker-library/go-dockerlibrary/architecture/oci-platform.go
-ARCH_MAP = { "x86_64":    "amd64",
-             "armv5l":    "arm/v5",
+ARCH_MAP = { "armv5l":    "arm/v5",
              "armv6l":    "arm/v6",
              "aarch32":   "arm/v7",
              "armv7l":    "arm/v7",
@@ -74,7 +91,8 @@ ARCH_MAP = { "x86_64":    "amd64",
              "i686":      "386",
              "mips64le":  "mips64le",
              "ppc64le":   "ppc64le",
-             "s390x":     "s390x" }  # a.k.a. IBM Z
+             "s390x":     "s390x",     # a.k.a. IBM Z
+             "x86_64":    "amd64" }
 
 # Some images have oddly specified architecture. For example, as of
 # 2022-06-08, on Docker Hub, opensuse/leap:15.1 offers architectures amd64,
@@ -101,8 +119,9 @@ FILENAME_MAX_CHARS = 192
 # takes over.
 HTTP_CHUNK_SIZE = 256 * 1024
 
-# Minimum Python version. NOTE: Keep in sync with configure.ac.
+# Minimum versions. NOTE: Keep in sync with configure.ac.
 PYTHON_MIN = (3,6)
+RSYNC_MIN = (3,1,0)
 
 
 ## Globals ##
@@ -120,10 +139,10 @@ CH_BIN = None
 CH_RUN = None
 
 # Logging; set using init() below.
-verbose = 0          # Verbosity level.
-log_festoon = False  # If true, prepend pid and timestamp to chatter.
-log_fp = sys.stderr  # File object to print logs to.
-trace_fatal = False  # Add abbreviated traceback to fatal error hint.
+log_level = Log_Level(0)  # Verbosity level.
+log_festoon = False       # If true, prepend pid and timestamp to chatter.
+log_fp = sys.stderr       # File object to print logs to.
+trace_fatal = False       # Add abbreviated traceback to fatal error hint.
 
 # Warnings to be re-printed when program exits
 warnings = list()
@@ -142,8 +161,8 @@ class Fatal_Error(Exception):
    def __init__(self, *args, **kwargs):
       self.args = args
       self.kwargs = kwargs
-class No_Fatman_Error(Exception): pass
 class Image_Unavailable_Error(Exception): pass
+class No_Fatman_Error(Exception): pass
 
 
 ## Classes ##
@@ -305,6 +324,11 @@ class Progress:
       self.display_last = float("-inf")
       self.update(0)
 
+   def done(self):
+      self.update(0, True)
+      if (self.overwrite_p):
+         INFO("")  # newline to release display line
+
    def update(self, increment, last=False):
       now = time.monotonic()
       self.progress += increment
@@ -326,11 +350,6 @@ class Progress:
          INFO(line, end=("\r" if self.overwrite_p else "\n"))
          self.display_last = now
 
-   def done(self):
-      self.update(0, True)
-      if (self.overwrite_p):
-         INFO("")  # newline to release display line
-
 
 class Progress_Reader:
    """Wrapper around a binary file object to maintain a progress meter while
@@ -379,26 +398,45 @@ class Progress_Reader:
 
 class Progress_Writer:
    """Wrapper around a binary file object to maintain a progress meter while
-      data are written."""
+      data are written. Overwrite the file if it already exists.
+
+      This downloads to a temporary file to ease recovery if the download is
+      interrupted. This uses a predictable name to support restarts in the
+      future, which would probably require verification after download. For
+      now, we just delete any leftover temporary files in Storage.init().
+
+      An interesting alternative is to download to an anonymous temporary file
+      that vanishes if not linked into the filesystem. Recent Linux provides a
+      very cool procedure to do this -- open(2) with O_TMPFILE followed by
+      linkat(2) [1] -- but it’s not always supported and the workaround
+      (create, then immediately unlink(2)) does not support re-linking [2].
+      This would also not support restarting the download.
+
+      [1]: https://man7.org/linux/man-pages/man2/open.2.html
+      [2]: https://stackoverflow.com/questions/4171713"""
 
    __slots__ = ("fp",
                 "msg",
                 "path",
+                "path_tmp",
                 "progress")
 
    def __init__(self, path, msg):
       self.msg = msg
       self.path = path
+      self.path_tmp = path.with_name("part_" + path.name)
       self.progress = None
 
    def close(self):
       if (self.progress is not None):
          self.progress.done()
          close_(self.fp)
+         self.path.unlink_(missing_ok=True)
+         self.path_tmp.rename_(self.path)
 
    def start(self, length):
       self.progress = Progress(self.msg, "MiB", 2**20, length)
-      self.fp = self.path.open_("wb")
+      self.fp = self.path_tmp.open_("wb")
 
    def write(self, data):
       self.progress.update(len(data))
@@ -419,7 +457,7 @@ class Timer:
 ## Supporting functions ##
 
 def DEBUG(msg, hint=None, **kwargs):
-   if (verbose >= 2):
+   if (log_level >= Log_Level.DEBUG):
       log(msg, hint, None, "38;5;6m", "", **kwargs)  # dark cyan (same as 36m)
 
 def ERROR(msg, hint=None, trace=None, **kwargs):
@@ -438,20 +476,22 @@ def FATAL(msg, hint=None, **kwargs):
 
 def INFO(msg, hint=None, **kwargs):
    "Note: Use print() for output; this function is for logging."
-   log(msg, hint, None, "33m", "", **kwargs)  # yellow
+   if (log_level >= Log_Level.INFO):
+      log(msg, hint, None, "33m", "", **kwargs)  # yellow
 
 def TRACE(msg, hint=None, **kwargs):
-   if (verbose >= 3):
+   if (log_level >= Log_Level.TRACE):
       log(msg, hint, None, "38;5;6m", "", **kwargs)  # dark cyan (same as 36m)
 
 def VERBOSE(msg, hint=None, **kwargs):
-   if (verbose >= 1):
+   if (log_level >= Log_Level.VERBOSE):
       log(msg, hint, None, "38;5;14m", "", **kwargs)  # light cyan (1;36m, not bold)
 
 def WARNING(msg, hint=None, msg_save=True, **kwargs):
-   if (msg_save):
-      warnings.append(msg)
-   log(msg, hint, None, "31m", "warning: ", **kwargs)  # red
+   if (log_level > Log_Level.STDERR):
+      if (msg_save):
+         warnings.append(msg)
+      log(msg, hint, None, "31m", "warning: ", **kwargs)  # red
 
 def arch_host_get():
    "Return the registry architecture of the host."
@@ -494,6 +534,10 @@ def cmd(argv, fail_ok=False, **kwargs):
    """Run command using cmd_base(). If fail_ok, return the exit code whether
       or not the process succeeded; otherwise, return (zero) only if the
       process succeeded and exit with fatal error if it failed."""
+   if (log_level < Log_Level.WARNING):
+      kwargs["stdout"] = subprocess.DEVNULL
+      if (log_level <= Log_Level.QUIET_STDERR):
+         kwargs["stderr"] = subprocess.DEVNULL
    cp = cmd_base(argv, fail_ok=fail_ok, **kwargs)
    return cp.returncode
 
@@ -510,7 +554,7 @@ def cmd_base(argv, fail_ok=False, **kwargs):
       for (k,v) in sorted(kwargs["env"].items()):
          VERBOSE("env: %s=%s" % (k,v))
    if ("stderr" not in kwargs):
-      if (verbose <= 1):  # VERBOSE or lower: capture for printing on fail only
+      if (log_level <= Log_Level.INFO):  # VERBOSE or lower: capture for printing on fail only
          kwargs["stderr"] = subprocess.PIPE
    if ("input" not in kwargs):
       kwargs["stdin"] = subprocess.DEVNULL
@@ -538,26 +582,26 @@ def cmd_base(argv, fail_ok=False, **kwargs):
             % (cp.returncode, argv_to_string(argv)))
    return cp
 
+def cmd_quiet(argv, **kwargs):
+   """Run command using cmd() and return the exit code. If logging is verbose
+      or lower, discard stdout."""
+   if (log_level >= Log_Level.DEBUG):  # debug or higher
+      stdout=None
+   else:
+      stdout=subprocess.DEVNULL
+   return cmd(argv, stdout=stdout, **kwargs)
+
 def cmd_stdout(argv, encoding="UTF-8", **kwargs):
    """Run command using cmd_base(), capturing its standard output. Return the
       CompletedProcess object (its stdout is available in the “stdout”
       attribute). If logging is debug or higher, print stdout."""
    cp = cmd_base(argv, encoding=encoding, stdout=subprocess.PIPE, **kwargs)
-   if (verbose >= 2):  # debug or higher
+   if (log_level >= Log_Level.DEBUG):  # debug or higher
       # just dump to stdout rather than using DEBUG() to match cmd_quiet
       sys.stdout.write(cp.stdout)
       sys.stdout.flush()
    return cp
 
-def cmd_quiet(argv, **kwargs):
-   """Run command using cmd() and return the exit code. If logging is verbose
-      or lower, discard stdout."""
-   if (verbose >= 2):  # debug or higher
-      stdout=None
-   else:
-      stdout=subprocess.DEVNULL
-   return cmd(argv, stdout=stdout, **kwargs)
-
 def color_reset(*fps):
    for fp in fps:
       color_set("0m", fp)
@@ -566,10 +610,6 @@ def color_set(color, fp):
    if (fp.isatty()):
       print("\033[" + color, end="", flush=True, file=fp)
 
-def copy2(src, dst, **kwargs):
-   "Wrapper for shutil.copy2() with error checking."
-   ossafe(shutil.copy2, "can’t copy: %s -> %s" % (src, dst), src, dst, **kwargs)
-
 def dependencies_check():
    """Check more dependencies. If any dependency problems found, here or above
       (e.g., lark module checked at import time), then complain and exit."""
@@ -613,20 +653,28 @@ def exit(code):
 
 def init(cli):
    # logging
-   global log_festoon, log_fp, trace_fatal, verbose, save_xattrs
-   assert (0 <= cli.verbose <= 3)
-   verbose = cli.verbose
+   global log_festoon, log_fp, log_level, trace_fatal, save_xattrs
    save_xattrs = (not cli.no_xattrs)
    trace_fatal = (cli.debug or bool(os.environ.get("CH_IMAGE_DEBUG", False)))
+   if (cli.quiet and cli.verbose):
+      ERROR("“--quiet” incompatible with “--verbose”")
+      FATAL("incompatible option")
+   log_level = Log_Level(cli.verbose - cli.quiet)
+   assert (-3 <= log_level.value <= 3)
+   if (log_level <= Log_Level.STDERR):
+      # suppress writing to stdout (particularly “print”).
+      sys.stdout = open(os.devnull, 'w')
    if ("CH_LOG_FESTOON" in os.environ):
       log_festoon = True
    file_ = os.getenv("CH_LOG_FILE")
    if (file_ is not None):
-      verbose = max(verbose, 1)
       log_fp = file_.open_("at")
    atexit.register(color_reset, log_fp)
    VERBOSE("version: %s" % version.VERSION)
-   VERBOSE("verbose level: %d" % verbose)
+   VERBOSE("verbose level: %d (%s))" % (log_level.value, log_level.name))
+   # signal handling
+   signal.signal(signal.SIGINT, sigterm)
+   signal.signal(signal.SIGTERM, sigterm)
    # storage directory
    global storage
    storage = fs.Storage(cli.storage)
@@ -689,18 +737,6 @@ def kill_blocking(pid, timeout=10):
    FATAL("timeout of %ds exceeded trying to kill PID %d" % (timeout, pid),
          BUG_REPORT_PLZ)
 
-def walk(*args, **kwargs):
-   """Wrapper for os.walk(). Return a generator of the files in a directory
-      tree (root specified in *args). For each directory in said tree, yield a
-      3-tuple (dirpath, dirnames, filenames), where dirpath is a Path object,
-      and dirnames and filenames are lists of Path objects. For insight into
-      these being lists rather than generators, see use of ch.walk() in
-      I_copy.copy_src_dir()."""
-   for (dirpath, dirnames, filenames) in os.walk(*args, **kwargs):
-      yield (fs.Path(dirpath),
-             [fs.Path(dirname) for dirname in dirnames],
-             [fs.Path(filename) for filename in filenames])
-
 def log(msg, hint, trace, color, prefix, end="\n"):
    if (color is not None):
       color_set(color, log_fp)
@@ -803,6 +839,17 @@ def si_decimal(ct):
       ct /= 1000
    assert False, "unreachable"
 
+def sigterm(signum, frame):
+   "Handler for SIGTERM and friends."
+   # Ignore further signals because we are already cleaning up.
+   signal.signal(signal.SIGINT, signal.SIG_IGN)
+   signal.signal(signal.SIGTERM, signal.SIG_IGN)
+   # Don’t stomp on progress meter if one is being printed.
+   print()
+   signame = signal.Signals(signum).name
+   ERROR("received %s, exiting" % signame)
+   FATAL("received %s" % signame)
+
 def user():
    "Return the current username; exit with error if it can’t be obtained."
    try:
@@ -860,6 +907,18 @@ def version_check(argv, min_, required=True, regex=r"(\d+)\.(\d+)\.(\d+)"):
    VERBOSE("%s version OK: %d.%d.%d ≥ %d.%d.%d" % ((prog,) + v + min_))
    return True
 
+def walk(*args, **kwargs):
+   """Wrapper for os.walk(). Return a generator of the files in a directory
+      tree (root specified in *args). For each directory in said tree, yield a
+      3-tuple (dirpath, dirnames, filenames), where dirpath is a Path object,
+      and dirnames and filenames are lists of Path objects. For insight into
+      these being lists rather than generators, see use of ch.walk() in
+      I_copy.copy_src_dir()."""
+   for (dirpath, dirnames, filenames) in os.walk(*args, **kwargs):
+      yield (fs.Path(dirpath),
+             [fs.Path(dirname) for dirname in dirnames],
+             [fs.Path(filename) for filename in filenames])
+
 def warnings_dump():
    if (len(warnings) > 0):
       WARNING("reprinting %d warning(s)" % len(warnings), msg_save=False)
