@@ -8,6 +8,7 @@ import re
 import pprint
 import shutil
 import stat
+import struct
 import tarfile
 
 import charliecloud as ch
@@ -32,49 +33,81 @@ STORAGE_VERSION = 7
 storage_lock = True
 
 
+### Functions ###
+
+def copy(src, dst, follow_symlinks=False):
+   """Copy file src to dst. Wrapper function providing same signature as
+      shutil.copy2(). See Path.copy() for lots of gory details. Accepts
+      follow_symlinks, but the only valid value is False."""
+   assert (not follow_symlinks)
+   if (isinstance(src, str)):
+      src = Path(src)
+   if (isinstance(dst, str)):
+      dst = Path(dst)
+   src.copy(dst)
+
+
 ## Classes ##
 
 class Path(pathlib.PosixPath):
-   """Stock Path objects have the very weird property that appending an
-      *absolute* path to an existing path ignores the left operand, leaving
-      only the absolute right operand:
+   """Filesystem paths with two important differences from the stock Path:
 
-        >>> import pathlib
-        >>> a = pathlib.Path("/foo/bar")
-        >>> a.joinpath("baz")
-        PosixPath('/foo/bar/baz')
-        >>> a.joinpath("/baz")
-        PosixPath('/baz')
+      1. This Path remembers whether a trailing slash is present, and appends
+         it when str() or repr(). Note that while Path("/") is considered
+         internally *not* to have a trailing slash, when stringified (as "/")
+         and passed to rsync(1) on its command line, rsync(1) *does* interpret
+         it as trailing-slashed.
 
-      This is contrary to long-standing UNIX/POSIX, where extra slashes in a
-      path are ignored, e.g. the path "foo//bar" is equivalent to "foo/bar".
-      It seems to be inherited from os.path.join().
+      2. Stock Path objects have the very weird property that appending an
+         absolute path to an existing path ignores the left operand, leaving
+         only the absolute right operand:
 
-      Even with the relatively limited use of Path objects so far, this has
-      caused quite a few bugs. IMO it’s too difficult and error-prone to
-      manually manage whether paths are absolute or relative. Thus, this
-      subclass introduces a new operator "//" which does the right thing,
-      i.e., if the right operand is absolute, that fact is ignored. E.g.:
+           >>> import pathlib
+           >>> a = pathlib.Path("/foo/bar")
+           >>> a.joinpath("baz")
+           PosixPath('/foo/bar/baz')
+           >>> a.joinpath("/baz")
+           PosixPath('/baz')
 
-        >>> a = Path("/foo/bar")
-        >>> a.joinpath_posix("baz")
-        Path('/foo/bar/baz')
-        >>> a.joinpath_posix("/baz")
-        Path('/foo/bar/baz')
-        >>> a // "/baz"
-        Path('/foo/bar/baz')
-        >>> "/baz" // a
-        Path('/baz/foo/bar')
+        This is contrary to long-standing UNIX/POSIX, where extra slashes in a
+        path are ignored, e.g. the path "foo//bar" is equivalent to "foo/bar".
+        It seems to be inherited from os.path.join().
 
-      We introduce a new operator because it seemed like too subtle a change
-      to the existing operator "/" (which we disable to avoid getting burned
-      here in Charliecloud). An alternative was "+" like strings, but that led
-      to silently wrong results when the paths *were* strings (components
-      concatenated with no slash)."""
+        Even with the relatively limited use of Path objects so far, this has
+        caused quite a few bugs. IMO it’s too difficult and error-prone to
+        manually manage whether paths are absolute or relative. Thus, this
+        subclass introduces a new operator "//" which does the right thing,
+        i.e., if the right operand is absolute, that fact is ignored. E.g.:
+
+           >>> a = Path("/foo/bar")
+           >>> a.joinpath_posix("baz")
+           Path('/foo/bar/baz')
+           >>> a.joinpath_posix("/baz")
+           Path('/foo/bar/baz')
+           >>> a // "/baz"
+           Path('/foo/bar/baz')
+           >>> "/baz" // a
+           Path('/baz/foo/bar')
+
+        We introduce a new operator because it seemed like too subtle a change
+        to the existing operator "/" (which we disable to avoid getting burned
+        here in Charliecloud). An alternative was "+" like strings, but that
+        led to silently wrong results when the paths *were* strings
+        (components concatenated with no slash)."""
+
+   __slots__ = ("trailing_slash_p",)
 
    # Name of the gzip(1) to use; set on first call of file_gzip().
    gzip = None
 
+   def __init__(self, *args):
+      # WARNING: Something very weird is going on with the superclass [1] that
+      # means super().__init__ refers to object.__init__, not
+      # PosixPath.__init__. All seems well if we don’t call it but omg.
+      # [1]: https://stackoverflow.com/questions/61689391
+      self.trailing_slash_p = (    len(args) > 0 and isinstance(args[-1], str)
+                               and len(args[-1]) > 1 and args[-1][-1] == "/")
+
    def __floordiv__(self, right):
       return self.joinpath_posix(right)
 
@@ -88,6 +121,13 @@ class Path(pathlib.PosixPath):
    def __rtruediv__(self, left):
       return NotImplemented
 
+   def __str__(self):
+      # Because subclassing pathlib.Path is so bonkers, we often get instances
+      # that don’t have attribute trailing_slash_p. Don’t crash in this case,
+      # and cross our fingers we haven’t just lost the trailing slash.
+      return super().__str__() + ("/" if (    hasattr(self, "trailing_slash_p")
+                                          and self.trailing_slash_p) else "")
+
    def __truediv__(self, right):
       return NotImplemented
 
@@ -111,6 +151,37 @@ class Path(pathlib.PosixPath):
       "Return True if I can’t be stored in Git because of my name."
       return self.name.startswith(".git")
 
+   @property
+   def is_root(self):
+      return (str(self) == "/")
+
+   @property
+   def mountpoint(self):
+      """Return the mount point of the filesystem containing me (or, if
+         symlink, the file pointed to)."""
+      # https://stackoverflow.com/a/4453715
+      try:
+         pc = self.resolve(strict=True)
+      except RuntimeError:
+         ch.FATAL("not found, can’t resolve: %s" % self)
+      # Unclear whether ismount() deals correctly with the root directory, so
+      # do the stat(2) stuff ourself.
+      dev_child = pc.stat().st_dev
+      while (not pc.is_root):
+         dev_parent = pc.parent.stat().st_dev
+         if (dev_child != dev_parent):
+            return pc
+         pc = pc.parent
+      # Got all the way up to root without finding a transition, so we’re on
+      # the root filesystem.
+      return Path("/")
+
+   @property
+   def parent(self):
+      ret = super().parent
+      ret.trailing_slash_p = False  # trailing slash in removed part
+      return ret
+
    @classmethod
    def gzip_set(cls):
       """Set gzip class attribute on first call to file_gzip().
@@ -128,6 +199,14 @@ class Path(pathlib.PosixPath):
          else:
             ch.FATAL("can’t find path to gzip or pigz")
 
+   @staticmethod
+   def stat_bytes_all(paths):
+      "Return concatenation of metadata_bytes() on each given Path object."
+      md = bytearray()
+      for path in paths:
+         md += path.stat_bytes_recursive()
+      return md
+
    def add_suffix(self, suff):
       """Returns the path object restulting from appending the specified
          suffix to the end of the path name. E.g. Path(foo).add_suffix(".txt")
@@ -162,9 +241,68 @@ class Path(pathlib.PosixPath):
          ch.ossafe(os.chmod, "can’t chmod: %s" % self, self, perms_new)
       return (st.st_mode | perms_new)
 
+   def copy(self, dst):
+      """Copy file myself to dst, including metadata, overwriting dst if it
+         exists. dst must be the actual destination path, i.e., it may not be
+         a directory. Does not follow symlinks.
+
+         If (a) src is a regular file, (b) src and dst are on the same
+         filesystem, and (c) Python is version ≥3.8, then use
+         os.copy_file_range() [1,2], which at a minimum does an in-kernel data
+         transfer. If that filesystem also (d) supports copy-on-write [3],
+         then this is a very fast lazy reflink copy.
+
+         [1]: https://docs.python.org/3/library/os.html#os.copy_file_range
+         [2]: https://man7.org/linux/man-pages/man2/copy_file_range.2.html
+         [3]: https://elixir.bootlin.com/linux/latest/A/ident/remap_file_range
+      """
+      src_st = self.stat_(False)
+      # dst is not a directory, so parent must be on the same filesystem. We
+      # *do* want to follow symlinks on the parent.
+      dst_dev = dst.parent.stat_(True).st_dev
+      if (    stat.S_ISREG(src_st.st_mode)
+          and src_st.st_dev == dst_dev
+          and hasattr(os, "copy_file_range")):
+         # Fast path. The same-filesystem restriction is because reliable
+         # copy_file_range(2) between filesystems seems quite new (maybe
+         # kernel 5.18?).
+         try:
+            if (dst.exists()):
+               # If dst is a symlink, we get OLOOP from os.open(). Delete it
+               # unconditionally though, for simplicity.
+               dst.unlink()
+            src_fd = os.open(self, os.O_RDONLY|os.O_NOFOLLOW)
+            dst_fd = os.open(dst, os.O_WRONLY|os.O_NOFOLLOW|os.O_CREAT)
+            # I’m not sure why we need to loop this -- there’s no explanation
+            # of *when* fewer bytes than requested would be copied -- but the
+            # man page example does.
+            remaining = src_st.st_size
+            while (remaining > 0):
+               copied = os.copy_file_range(src_fd, dst_fd, remaining)
+               if (copied == 0):
+                  ch.FATAL("zero bytes copied: %s -> %s" % (self, dst))
+               remaining -= copied
+            os.close(src_fd)
+            os.close(dst_fd)
+         except OSError as x:
+            ch.FATAL("can’t copy data (fast): %s -> %s: %s"
+                     % (self, dst, x.strerror))
+      else:
+         # Slow path.
+         try:
+            shutil.copyfile(self, dst, follow_symlinks=False)
+         except OSError as x:
+            ch.FATAL("can’t copy data (slow): %s -> %s: %s"
+                     % (self, dst, x.strerror))
+      try:
+         # Metadata.
+         shutil.copystat(self, dst, follow_symlinks=False)
+      except OSError as x:
+         ch.FATAL("can’t copy metadata: %s -> %s" % (self, dst, x.strerror))
+
    def copytree(self, *args, **kwargs):
       "Wrapper for shutil.copytree() that exits on the first error."
-      shutil.copytree(str(self), copy_function=ch.copy2, *args, **kwargs)
+      shutil.copytree(self, copy_function=copy, *args, **kwargs)
 
    def disk_bytes(self):
       """Return the number of disk bytes consumed by path. Note this is
@@ -298,22 +436,32 @@ class Path(pathlib.PosixPath):
          except ValueError:
             return False
 
-   def joinpath_posix(self, other):
+   def joinpath_posix(self, right):
       # This method is a hot spot, so the hairiness is due to optimizations.
-      # It runs about 30% faster than the naïve verson below.
-      if (isinstance(other, Path)):
-         other_parts = other._parts
-         if (len(other_parts) > 0 and other_parts[0] == "/"):
-            other_parts = other_parts[1:]
-      elif (isinstance(other, str)):
-         other_parts = other.split("/")
-         if (len(other_parts) > 0 and len(other_parts[0]) == 0):
-            other_parts = other_parts[1:]
+      # It runs about 30% faster than the naïve version referenced below.
+      if (isinstance(right, Path)):
+         right_parts = right._parts
+         ts_p = right.trailing_slash_p
+         if (len(right_parts) > 0 and right_parts[0] == "/"):
+            right_parts = right_parts[1:]
+      elif (isinstance(right, str)):
+         right_parts = right.split("/")
+         if (len(right_parts) > 0 and len(right_parts[0]) == 0):
+            # absolute path, rm empty first component
+            right_parts = right_parts[1:]
+         if (len(right_parts) > 0 and len(right_parts[-1]) == 0):
+            # trailing slash, rm empty last component
+            right_parts = right_parts[:-1]
+            ts_p = True
+         else:
+            ts_p = False
       else:
-         ch.INFO(type(other))
-         assert False, "unknown type"
-      return self._from_parsed_parts(self._drv, self._root,
-                                     self._parts + other_parts)
+         assert False, "unknown type: %s" % type(right)
+      ret = self._from_parsed_parts(self._drv, self._root,
+                                    self._parts + right_parts)
+      ret.trailing_slash_p = ts_p
+      return ret
+
       # Naïve implementation for reference.
       #other = Path(other)
       #if (other.is_absolute()):
@@ -339,22 +487,6 @@ class Path(pathlib.PosixPath):
          effort required to make the change."""
       return set(ch.ossafe(os.listdir, "can’t list: %s" % self.name, self))
 
-   def strip(self, left=0, right=0):
-      """Return a copy of myself with n leading components removed. E.g.:
-
-           >>> a = Path("/a/b/c")
-           >>> a.strip(left=1)
-           Path("a/b/c")
-           >>> a.strip(right=1)
-           Path("/a/b")
-           >>> a.strip(left=1, right=1)
-           Path("a/b")
-
-         It is an error if I don’t have at least left + right components,
-         i.e., you can strip a path down to nothing but not further."""
-      assert (len(self.parts) >= left + right)
-      return Path(*self.parts[left:len(self.parts)-right])
-
    def mkdir_(self):
       ch.TRACE("ensuring directory: %s" % self)
       try:
@@ -378,6 +510,11 @@ class Path(pathlib.PosixPath):
                        "can’t open for %s: %s" % (mode, self.name),
                        mode, *args, **kwargs)
 
+   def relative_to(self, other):  # FIXME: does not support component-wise
+      ret = super().relative_to(other)
+      ret.trailing_slash_p = other.trailing_slash_p
+      return ret
+
    def rename_(self, name_new):
       if (Path(name_new).exists()):
          ch.FATAL("can’t rename: destination exists: %s" % name_new)
@@ -385,6 +522,11 @@ class Path(pathlib.PosixPath):
                 "can’t rename: %s -> %s" % (self.name, name_new),
                 name_new)
 
+   def resolve(self, *args, **kwargs):
+      ret = super().resolve(*args, **kwargs)
+      ret.trailing_slash_p = self.trailing_slash_p
+      return ret
+
    def rmdir_(self):
       ch.ossafe(super().rmdir, "can’t rmdir: %s" % self.name)
 
@@ -399,6 +541,22 @@ class Path(pathlib.PosixPath):
       else:
          assert False, "unimplemented"
 
+   def setxattr(self, name, value, follow_symlinks=True):
+      if (ch.save_xattrs):
+         try:
+            os.setxattr(self, name, value, follow_symlinks)
+         except OSError as x:
+            if (x.errno == errno.ENOTSUP):  # no OSError subclass
+               ch.WARNING("xattrs not supported on %s, setting --no-xattr"
+                          % (self.mountpoint))
+               ch.save_xattrs = False
+            else:
+               ch.FATAL("can’t set xattr: %s: %s: %s"
+                        % (self, name, x.strerror))
+      if (not ch.save_xattrs):  # not “else” because could have change in “if”
+         ch.DEBUG("xattrs disabled, ignoring: %s: %s" % (self, name))
+         return
+
    def stat_(self, links):
       """An error-checking version of stat(). Note that we cannot simply
          change the definition of stat() to be ossafe, as the exists() method
@@ -410,7 +568,42 @@ class Path(pathlib.PosixPath):
          follow_symlinks kwarg is absent in pathlib for Python 3.6, which we
          want to retain compatibility with."""
       return ch.ossafe(os.stat, "can’t stat: %s" % self, self,
-                    follow_symlinks=links)
+                       follow_symlinks=links)
+
+   def stat_bytes(self, links=False):
+      "Return self.stat_() encoded as an opaque bytearray."
+      st = self.stat_(links)
+      return (  str(self).encode("UTF-8")
+              + struct.pack("=HQQ", st.st_mode, st.st_size, st.st_mtime_ns))
+
+   def stat_bytes_recursive(self):
+      """Return concatenation of self.stat_() and all my children as an opaque
+         bytearray, in unspecified but consistent order."""
+      # FIXME: Locale issues related to sorting?
+      md = self.stat_bytes(links=True)
+      if (self.is_dir()):
+         for (dir_, dirs, files) in ch.walk(self):
+            md += dir_.stat_bytes()
+            for f in sorted(files):
+               md += (dir_ // f).stat_bytes()
+            dirs.sort()
+      return md
+
+   def strip(self, left=0, right=0):
+      """Return a copy of myself with n leading components removed. E.g.:
+
+           >>> a = Path("/a/b/c")
+           >>> a.strip(left=1)
+           Path("a/b/c")
+           >>> a.strip(right=1)
+           Path("/a/b")
+           >>> a.strip(left=1, right=1)
+           Path("a/b")
+
+         It is an error if I don’t have at least left + right components,
+         i.e., you can strip a path down to nothing but not further."""
+      assert (len(self.parts) >= left + right)
+      return Path(*self.parts[left:len(self.parts)-right])
 
    def symlink(self, target, clobber=False):
       if (clobber and self.is_file()):
@@ -522,6 +715,20 @@ class Storage:
    def build_large_path(self, name):
       return self.build_large // name
 
+   def cleanup(self):
+      "Called during initialization after we know the storage dir is valid."
+      # Delete partial downloads.
+      part_ct = 0
+      for path in self.download_cache.glob("part_*"):
+         ch.VERBOSE("deleting: %s" % path)
+         path.unlink_()
+         part_ct += 1
+      if (part_ct > 0):
+         ch.WARNING("deleted %d partially downloaded files" % part_ct)
+
+   def fatman_for_download(self, image_ref):
+      return self.download_cache // ("%s.fat.json" % image_ref.for_path)
+
    def init(self):
       """Ensure the storage directory exists, contains all the appropriate
          top-level directories & metadata, and is the appropriate version."""
@@ -590,6 +797,7 @@ class Storage:
                   % (v_found, self.root),
                   'you can delete and re-initialize with "ch-image reset"')
       self.validate_strict()
+      self.cleanup()
 
    def lock(self):
       """Lock the storage directory. Charliecloud does not at present support
@@ -620,9 +828,6 @@ class Storage:
       return (   self.download_cache
               // ("%s%%%s.manifest.json" % (image_ref.for_path, digest)))
 
-   def fatman_for_download(self, image_ref):
-      return self.download_cache // ("%s.fat.json" % image_ref.for_path)
-
    def reset(self):
       if (self.valid_p):
          self.root.rmtree()
@@ -655,7 +860,15 @@ class Storage:
             entries.remove(entry)
          except KeyError:
             ch.FATAL("%s: missing file or directory: %s" % (msg_prefix, entry))
+      # Ignore some files that may or may not exist.
       entries -= { i.name for i in (self.lockfile, self.mount_point) }
+      # Delete some files that exist only if we crashed.
+      for i in (self.image_tmp, ):
+         if (i.name in entries):
+            ch.WARNING("deleting leftover temporary file/dir: %s" % i.name)
+            i.rmtree()
+            entries.remove(i.name)
+      # If anything is left, yell about it.
       if (len(entries) > 0):
          ch.FATAL("%s: extraneous file(s): %s"
                % (msg_prefix, " ".join(sorted(entries))))
