@@ -13,12 +13,12 @@ def main(cli):
    src_ref = im.Reference(cli.source_ref)
    ch.INFO("pushing image:   %s" % src_ref)
    image = im.Image(src_ref, cli.image)
-   # FIXME: validate it's an image using Megan's new function (PR #908)
+   # FIXME: validate it’s an image using Megan’s new function (PR #908)
    if (not os.path.isdir(image.unpack_path)):
       if (cli.image is not None):
-         ch.FATAL("can't push: %s does not appear to be an image" % cli.image)
+         ch.FATAL("can’t push: %s does not appear to be an image" % cli.image)
       else:
-         ch.FATAL("can't push: no image %s" % src_ref)
+         ch.FATAL("can’t push: no image %s" % src_ref)
    if (cli.image is not None):
       ch.INFO("image path:      %s" % image.unpack_path)
    else:
@@ -44,7 +44,8 @@ class Image_Pusher:
                 "dst_ref",   # destination of upload
                 "image",     # Image object we are uploading
                 "layers",    # list of (digest, .tar.gz path), lowest first
-                "manifest")  # sequence of bytes
+                "manifest",  # sequence of bytes
+                "registry")  # destination registry
 
    def __init__(self, image, dst_ref):
       self.config = None
@@ -52,6 +53,7 @@ class Image_Pusher:
       self.image = image
       self.layers = None
       self.manifest = None
+      self.registry = None
 
    @classmethod
    def config_new(class_):
@@ -82,7 +84,7 @@ class Image_Pusher:
 
    def cleanup(self):
       ch.INFO("cleaning up")
-      # Delete the tarballs since we can't yet cache them.
+      # Delete the tarballs since we can’t yet cache them.
       for (_, tar_c) in self.layers:
          ch.VERBOSE("deleting tarball: %s" % tar_c)
          tar_c.unlink_()
@@ -93,8 +95,13 @@ class Image_Pusher:
          manifest as a sequence of bytes).
 
          There is not currently any support for re-using any previously
-         prepared files already in the upload cache, because we don't yet have
+         prepared files already in the upload cache, because we don’t yet have
          a way to know if these have changed until they are already build."""
+      # Initializing an HTTP instance for the registry and doing a 'GET'
+      # request right out the gate ensures the user needs to authenticate
+      # before we prepare the image for upload (#1426).
+      self.registry = rg.HTTP(self.dst_ref)
+      self.registry.request("GET", self.registry._url_base)
       tars_uc = self.image.tarballs_write(ch.storage.upload_cache)
       tars_c = list()
       config = self.config_new()
@@ -125,8 +132,8 @@ class Image_Pusher:
                                  for k, v
                                  in self.image.metadata.get("env", {}).items()]
       # History. Some registries, e.g., Quay, use history metadata for simple
-      # sanity checks. For example, when an image's number of "empty_layer"
-      # history entries doesn't match the number of layers being uploaded,
+      # sanity checks. For example, when an image’s number of "empty_layer"
+      # history entries doesn’t match the number of layers being uploaded,
       # Quay will reject the image upload.
       #
       # This type of error checking is odd as the empty_layer key is optional
@@ -141,7 +148,7 @@ class Image_Pusher:
       for i in range(len(hist) - 1, -1, -1):
          if (   "empty_layer" not in hist[i].keys()
              or (    "empty_layer" in hist[i].keys()
-                 and not hist[i]["empty_layer"] == True)):
+                 and not hist[i]["empty_layer"])):
             non_empty_winner = i
             break
       assert(non_empty_winner is not None)
@@ -169,10 +176,9 @@ class Image_Pusher:
 
    def upload(self):
       ch.INFO("starting upload")
-      ul = rg.HTTP(self.dst_ref)
       for (i, (digest, tarball)) in enumerate(self.layers, start=1):
-         ul.layer_from_file(digest, tarball,
-                            "layer %d/%d: " % (i, len(self.layers)))
-      ul.config_upload(self.config)
-      ul.manifest_upload(self.manifest)
-      ul.close()
+         self.registry.layer_from_file(digest, tarball,
+                                 "layer %d/%d: " % (i, len(self.layers)))
+      self.registry.config_upload(self.config)
+      self.registry.manifest_upload(self.manifest)
+      self.registry.close()
